"""Django models for the jugemaj app."""
from enum import IntEnum
from functools import cmp_to_key

from django.conf import settings
from django.contrib.contenttypes.fields import GenericForeignKey
from django.contrib.contenttypes.models import ContentType
from django.db import models
from django.urls import reverse

from ndh.models import NamedModel, TimeStampedModel
from ndh.utils import enum_to_choices

CHOICES = IntEnum("choix", "Super Bien OK Passable Insuffisant Nul")


def sort_candidates(a, b):
    """
    Sort candidates by their majority gauge.

    a & b are the majority gauges of the candidates A & B
    (p_A, α_A, q_A) & (p_B, α_B, q_B)
    where
    - α is the majority grade;
    - p the percentage of votes strictly above α;
    - q the percentage of votes strictly below α.
    thus, A > B ⇔ α_A > α_B || (α_A == α_B && (p_A > max(q_A, p_B, q_B) || q_B > max(p_A, q_A, p_B)))
    (eq. 2, p.11 of the second article in README.md)
    """
    a, b = a.majority_gauge(), b.majority_gauge()
    if a[1] > b[1]:
        return 1
    if b[1] > a[1]:
        return -1
    if a[0] > max(a[2], b[0], b[2]):
        return 1
    if b[0] > max(b[2], a[0], a[2]):
        return -1
    if b[2] > max(a[0], a[2], b[0]):
        return 1
    if a[2] > max(b[0], b[2], a[0]):
        return -1
    return 0


class Election(NamedModel, TimeStampedModel):
    """Main model for the jugemaj app."""
    description = models.TextField()
    creator = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE)
    end = models.DateTimeField("fin")
    ended = models.BooleanField("fini", default=False)

    def get_absolute_url(self):
        """Reverse to the detail view for this instance."""
        return reverse('jugemaj:election', kwargs={'slug': self.slug})

    def results(self):
        """Get the sorted list of all Candidates for this Election."""
        return sorted(self.candidate_set.all(), key=cmp_to_key(sort_candidates))


class Candidate(models.Model):
    """An Election as Candidate as choices."""
    election = models.ForeignKey(Election, on_delete=models.CASCADE)
    content_type = models.ForeignKey(ContentType, on_delete=models.CASCADE)
    object_id = models.PositiveIntegerField()
    content_object = GenericForeignKey('content_type', 'object_id')

    def __str__(self):
        """Print the candidate."""
        return str(self.content_object)

    def get_absolute_url(self):
        """Get the candidate's Election URL."""
        return self.election.get_absolute_url()

    def majority_gauge(self):
        """Compute the majority gauge of this Candidate."""
        count = self.vote_set.count()
        if not count:
            return (0, 10, 0)
        mention = self.vote_set.order_by('choice')[count // 2].choice
        if mention is None:
            print(self.vote_set.order_by('choice'))
            mention = 6
        return (self.vote_set.filter(choice__gt=mention).count() / count, mention,
                self.vote_set.filter(choice__lt=mention).count() / count)

    def votes(self):
        """Get the list of the votes for this Candidate."""
        count = self.vote_set.count()
        if count:
            return [self.vote_set.filter(choice=i).count() * 100 / count for i in CHOICES]
        return [0] * len(CHOICES)


class NamedCandidate(NamedModel, TimeStampedModel):
    """A common model to be used as a Candidate object."""
    def get_absolute_url(self):
        """TODO."""
        return reverse('jugemaj:candidate', kwargs={'slug': self.slug})


class Vote(TimeStampedModel):
    """In an Election, each elector can give a rating to each Candidate."""
    elector = models.ForeignKey(settings.AUTH_USER_MODEL, on_delete=models.CASCADE, null=True)
    candidate = models.ForeignKey(Candidate, on_delete=models.CASCADE)
    choice = models.IntegerField("choix", choices=enum_to_choices(CHOICES), null=True)

    class Meta:
        """An elector has only one rating on each Candidate."""
        unique_together = ('elector', 'candidate')

    def __str__(self):
        """Describe an elector rating."""
        return f'vote de {self.elector} pour {self.candidate}: {self.choice}'



"""Django tests for the jugemaj app."""
from django.contrib.auth.models import User
from django.test import TestCase
from django.urls import reverse

from .models import Candidate, Election


class JugeMajTests(TestCase):
    """Mais class for the tests."""
    def setUp(self):
        """Create a bunch of users for the other tests."""
        a, b, c = (User.objects.create_user(guy, email='%s@example.org' % guy, password=guy) for guy in 'abc')
        a.is_superuser = True
        a.save()

    def test_views(self):
        """Test the django views in the jugemaj app."""
        self.assertEqual(self.client.get(reverse('jugemaj:elections')).status_code, 200)
        self.assertEqual(self.client.get(reverse('jugemaj:create_election')).status_code, 302)
        self.client.login(username='a', password='a')
        self.assertEqual(self.client.get(reverse('jugemaj:create_election')).status_code, 200)
        r = self.client.post(
            reverse('jugemaj:create_election'), {
                'name': 'Élection du roi de Vénus',
                'description': 'Vénus n’a plus de roi, qui pensez-vous des candidats suivants ?',
                'end_0': '2025-03-09',
                'end_1': '03:19:45',
            })
        self.assertEqual(r.status_code, 302)
        self.assertEqual(Election.objects.count(), 1)
        slug = Election.objects.first().slug
        self.assertEqual(r.url, f'/election/{slug}')
        self.assertEqual(self.client.get(r.url).status_code, 200)
        self.assertEqual(self.client.get(reverse('jugemaj:election', kwargs={'slug': slug})).status_code, 200)
        self.client.logout()
        self.assertEqual(self.client.get(r.url).status_code, 200)
        self.assertEqual(self.client.get(reverse('jugemaj:create_candidate', kwargs={'slug': slug})).status_code, 302)
        self.client.login(username='b', password='b')
        self.assertEqual(self.client.get(reverse('jugemaj:create_candidate', kwargs={'slug': slug})).status_code, 200)
        r = self.client.post(reverse('jugemaj:create_candidate', kwargs={'slug': slug}), {'name': 'Capitaine Zorg'})
        self.assertEqual(r.status_code, 302)
        self.assertEqual(r.url, f'/election/{slug}')
        self.assertEqual(Candidate.objects.count(), 1)
        for name in ['Buzz l’Éclair', 'Timon et Pumba', 'Sénateur Palpatine', 'Aragorn', 'Totoro', 'Obi-Wan Kenobi']:
            r = self.client.post(reverse('jugemaj:create_candidate', kwargs={'slug': slug}), {'name': name})
        self.client.logout()
        url = reverse('jugemaj:vote', kwargs={'slug': slug})
        self.assertEqual(self.client.get(url).status_code, 302)
        self.client.login(username='b', password='b')
        r = self.client.get(url)
        self.assertEqual(r.status_code, 200)
        data = {'csrf_token': r.context['csrf_token']}
        management_form = r.context['formset'].management_form
        for i in 'TOTAL_FORMS', 'INITIAL_FORMS', 'MIN_NUM_FORMS', 'MAX_NUM_FORMS':
            data['%s-%s' % (management_form.prefix, i)] = management_form[i].value()
        for i in range(r.context['formset'].total_form_count()):
            form = r.context['formset'].forms[i]
            for field in form.fields:
                data[f'{form.prefix}-{field}'] = i % 6 + 1

        self.assertEqual(self.client.post(url, data).status_code, 302)
        self.assertEqual(self.client.get(reverse('jugemaj:election', kwargs={'slug': slug})).status_code, 200)



"""Register jugemaj models in django admin."""
from django.contrib.admin import site

from .models import Candidate, Election, Vote

for model in [Election, Candidate, Vote]:
    site.register(model)



"""Django forms for the jugemaj app."""
from django import forms

from ndh.forms import AccessibleDateTimeField

from .models import Election, Vote

VoteFormSet = forms.modelformset_factory(Vote,
                                         fields=(
                                             'candidate',
                                             'choice',
                                         ),
                                         extra=0,
                                         widgets={
                                             'choice': forms.RadioSelect,
                                             'candidate': forms.HiddenInput
                                         })


class ElectionForm(forms.ModelForm):
    """Form for the Election model."""
    end = AccessibleDateTimeField()

    class Meta:
        """Declare the model and fields for this form."""
        model = Election
        fields = ("name", "description", "end")



# Generated by Django 2.0.1 on 2018-01-28 20:14

import django.db.models.deletion
from django.conf import settings
from django.db import migrations, models

import autoslug.fields  # type: ignore


class Migration(migrations.Migration):

    initial = True

    dependencies = [
        migrations.swappable_dependency(settings.AUTH_USER_MODEL),
        ('contenttypes', '0002_remove_content_type_name'),
    ]

    operations = [
        migrations.CreateModel(
            name='Candidate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('object_id', models.PositiveIntegerField()),
                ('content_type',
                 models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='contenttypes.ContentType')),
            ],
        ),
        migrations.CreateModel(
            name='Election',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200, unique=True)),
                ('slug', autoslug.fields.AutoSlugField(editable=False, populate_from='name', unique=True)),
                ('description', models.TextField()),
                ('end', models.DateTimeField(verbose_name='fin')),
                ('ended', models.BooleanField(default=False, verbose_name='fini')),
                ('creator', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE,
                                              to=settings.AUTH_USER_MODEL)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='NamedCandidate',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('name', models.CharField(max_length=200, unique=True)),
                ('slug', autoslug.fields.AutoSlugField(editable=False, populate_from='name', unique=True)),
            ],
            options={
                'abstract': False,
            },
        ),
        migrations.CreateModel(
            name='Vote',
            fields=[
                ('id', models.AutoField(auto_created=True, primary_key=True, serialize=False, verbose_name='ID')),
                ('created', models.DateTimeField(auto_now_add=True)),
                ('updated', models.DateTimeField(auto_now=True)),
                ('choice',
                 models.IntegerField(choices=[(1, 'Super'), (2, 'Bien'), (3, 'OK'), (4, 'Passable'),
                                              (5, 'Insuffisant'), (6, 'Nul')],
                                     null=True,
                                     verbose_name='choix')),
                ('candidate', models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='jugemaj.Candidate')),
                ('elector',
                 models.ForeignKey(null=True, on_delete=django.db.models.deletion.CASCADE,
                                   to=settings.AUTH_USER_MODEL)),
            ],
        ),
        migrations.AddField(
            model_name='candidate',
            name='election',
            field=models.ForeignKey(on_delete=django.db.models.deletion.CASCADE, to='jugemaj.Election'),
        ),
        migrations.AlterUniqueTogether(
            name='vote',
            unique_together={('elector', 'candidate')},
        ),
    ]



"""Templatetags for django-jugemaj."""
from django import template

register = template.Library()


@register.filter
def candidate(candidate, candidates):
    """Get a candidate from the dict."""
    return candidates[candidate]



"""Django URLs for the jugemaj app."""
from django.urls import path

from .views import CandidateCreateView, ElectionCreateView, ElectionDetailView, ElectionListView, vote

app_name = 'jugemaj'
urlpatterns = [
    path('', ElectionListView.as_view(), name='elections'),
    path('election', ElectionCreateView.as_view(), name='create_election'),
    path('election/<slug:slug>', ElectionDetailView.as_view(), name='election'),
    path('election/<slug:slug>/candidate', CandidateCreateView.as_view(), name='create_candidate'),
    path('election/<slug:slug>/vote', vote, name='vote'),
    path('candidate/<slug:slug>', vote, name='candidate'),
]



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['jugemaj', 'jugemaj.migrations', 'jugemaj.templatetags']

package_data = \
{'': ['*'], 'jugemaj': ['templates/*', 'templates/jugemaj/*']}

install_requires = \
['ndh>=3.10.4,<4.0.0']

setup_kwargs = {
    'name': 'django-jugemaj',
    'version': '2.0.0',
    'description': 'A Django app for the Majority Judgment voting system',
    'long_description': '# Jugement Majoritaire\n[![Build Status](https://travis-ci.org/nim65s/django-jugemaj.svg?branch=master)](https://travis-ci.org/nim65s/django-jugemaj)\n[![Coverage Status](https://coveralls.io/repos/github/nim65s/django-jugemaj/badge.svg?branch=master)](https://coveralls.io/github/nim65s/django-jugemaj?branch=master)\n\n\nApplication de vote par [Jugement Majoritaire](https://fr.wikipedia.org/wiki/Jugement_majoritaire).\n\n\n## Requirements\n\n- [ndh](https://pypi.python.org/pypi/ndh) (which requires [django](https://www.djangoproject.com/),\n  [django-autoslug](https://github.com/justinmayer/django-autoslug/),\n  [django-bootstrap4](https://github.com/zostera/django-bootstrap4) (can be made optional on request))\n\nTested with:\n- Python 3.6, 3.7, 3.8\n- [Django](https://www.djangoproject.com/) 2.0, 2.1, 2.2, 3.0\n\n## Theory\n\n- [A theory of measuring, electing, and ranking. *Michel Balinski and Rida Laraki*. In PNAS 2007](https://doi.org/10.1073/pnas.0702634104)\n- [Majority Judgment vs Majority Rule, *Michel Balinski and Rida Laraki*](http://www.lamsade.dauphine.fr/sites/default/IMG/pdf/cahier_377.pdf)\n',
    'author': 'Guilhem Saurel',
    'author_email': 'guilhem.saurel@laas.fr',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://github.com/nim65s/django-jugemaj',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.6.1,<4.0.0',
}


setup(**setup_kwargs)



"""Views for django-jugemaj."""
from django.contrib import messages
from django.contrib.auth.decorators import login_required
from django.contrib.auth.mixins import LoginRequiredMixin
from django.shortcuts import get_object_or_404, redirect, render
from django.views.generic import CreateView, DetailView, ListView

from ndh.mixins import SuperUserRequiredMixin

from .forms import ElectionForm, VoteFormSet
from .models import Candidate, Election, NamedCandidate, Vote


class ElectionDetailView(DetailView):
    """View to detail an Election."""
    model = Election


class ElectionListView(ListView):
    """View to list Elections."""
    model = Election


class ElectionCreateView(SuperUserRequiredMixin, CreateView):
    """View to create an Election."""
    model = Election
    form_class = ElectionForm

    def form_valid(self, form):
        """Set the election creator."""
        form.instance.creator = self.request.user
        return super().form_valid(form)


class CandidateCreateView(LoginRequiredMixin, CreateView):
    """View to create a Candidate to an Election."""
    model = NamedCandidate
    fields = ("name", )

    def form_valid(self, form):
        """Set the right Election for this Candidate."""
        ret = super().form_valid(form)
        election = get_object_or_404(Election, slug=self.kwargs.get("slug"))
        Candidate.objects.create(election=election, content_object=form.instance)
        return ret

    def get_success_url(self):
        """Redirect to the Election page."""
        return get_object_or_404(Election, slug=self.kwargs.get("slug")).get_absolute_url()


@login_required
def vote(request, slug):
    """View to let a user vote."""
    election = get_object_or_404(Election, slug=slug)
    candidates = {}
    for candidate in election.candidate_set.all():
        Vote.objects.get_or_create(elector=request.user, candidate=candidate)
        candidates[candidate.pk] = candidate.content_object.name
    votes = Vote.objects.filter(elector=request.user, candidate__election=election)
    if request.method == "POST":
        formset = VoteFormSet(request.POST, queryset=votes)
        if formset.is_valid():
            formset.save()
            messages.success(request, "Votre vote a été enregistré")
            return redirect(election)
        messages.error(request, "Corrigez les erreurs")
    else:
        formset = VoteFormSet(queryset=votes)
    return render(request, "jugemaj/vote.html", {"formset": formset, "candidates": candidates})
